@@ -18,15 +18,18 @@ if __name__ == '__main__':
     r = Zeroconf()
     print("1. Testing registration of a service...")
     desc = {'version': '0.10', 'a': 'test value', 'b': 'another value'}
-    info = ServiceInfo("_http._tcp.local.",
-                       "My Service Name._http._tcp.local.",
-                       socket.inet_aton("127.0.0.1"), 1234, 0, 0, desc)
+    info = ServiceInfo(
+        "_http._tcp.local.",
+        "My Service Name._http._tcp.local.",
+        addresses=[socket.inet_aton("127.0.0.1")],
+        port=1234,
+        properties=desc,
+    )
     print("   Registering service...")
     r.register_service(info)
     print("   Registration done.")
     print("2. Testing query of service information...")
-    print("   Getting ZOE service: %s" % (
-        r.get_service_info("_http._tcp.local.", "ZOE._http._tcp.local.")))
+    print("   Getting ZOE service: %s" % (r.get_service_info("_http._tcp.local.", "ZOE._http._tcp.local.")))
     print("   Query done.")
     print("3. Testing query of own service...")
     queried_info = r.get_service_info("_http._tcp.local.", "My Service Name._http._tcp.local.")
